# 三、微分与积分

在本章中，我们将掌握一些经典的和最先进的技术，以执行微积分（以及扩展到物理和每个工程领域）的两个核心操作：功能的微分和集成。

# 动机

铁路或道路建筑（尤其是高速公路出口）以及许多过山车中那些疯狂的环路的设计共同点是二维或三维微分方程的解决方案，它解决了曲率和向心加速度对运动物体的影响 。 1970 年代，沃纳·斯坦格尔（Werner Stengel）研究并应用了多种模型来解决该问题，在他发现的众多解决方案中，有一种解决方案特别出色-使用回旋环（基于 Cornu 螺旋线的截面）。 1976 年，在美国加利福尼亚州巴伦西亚的六旗魔术山的 Baja Ridge 地区建造了第一个以此范式设计的环形过山车。 它是由美国大革命（Great American Revolution）创造的，它具有第一个垂直环（与两个开瓶器一起，总共三个倒置）。

![Motivation](img/image00392.jpeg)

设计中最棘手的部分是基于微分方程组，其解决方案取决于菲涅耳型正弦和余弦积分的积分，然后选择结果曲线的适当部分。 让我们看看这些有趣的函数的计算和绘图：

```py
In [1]: import numpy as np, matplotlib.pyplot as plt; \
 ...: from scipy.special import fresnel
In [2]: np.info(fresnel)
fresnel(x[, out1, out2])
(ssa,cca)=fresnel(z) returns the Fresnel sin and cos integrals:
integral(sin(pi/2 * t**2),t=0..z) and
integral(cos(pi/2 * t**2),t=0..z)
for real or complex z.
In [3]: ssa, cca = fresnel(np.linspace(-4, 4, 1000))
In [4]: plt.plot(ssa, cca, 'b-'); \
 ...: plt.axes().set_aspect('equal'); \
 ...: plt.show()

```

结果如下图：

![Motivation](img/image00393.jpeg)

菲涅耳积分的重要性使它们在 SciPy 库中具有永久性的地位。 还有许多其他有用的积分具有相同的命运，并且现在可以在模块`scipy.special`中进行操作了。 有关所有这些积分的完整列表以及其他相关功能的实现及其根源或派生，请参阅[上`scipy.special`的在线文档，网址为 http://docs.scipy.org/doc/scipy- 0.13.0 / reference / special.html](http://docs.scipy.org/doc/scipy-0.13.0/reference/special.html) ，或在第 4 章，*数值分析科学*中，Francisco Blanco-Silva 的*学习数值和科学计算科学[* 。

对于所有其他未包含在此函数列表中的函数，我们仍然需要可靠的解决方案来计算其根，导数或积分。 在本章中，我们将重点介绍允许执行最后两个操作的计算设备。

### 提示

下一章将介绍任何给定函数的根的计算（或近似）。

# 差异化

有三种方法可以计算导数：

*   数值微分是指某个点上给定函数的导数的近似过程。 在 SciPy 中，我们有以下过程，将详细介绍：
    *   对于一般的单变量函数，中心差公式具有固定的间距。
    *   始终可以通过柯西定理执行数值微分，将导数转换为定积分。 然后使用即将在后面一节中解释的数值积分技术来处理该积分。
*   符号区分是指函数派生函数的函数表达式的计算，几乎与我们手动进行的方式相同。 之所以称为符号，是因为符号与其数字对应物不同，它承担的是变量的作用，而不是数字或数字的向量。 要执行符号区分，我们需要一个**计算机代数系统**（ **CAS** ），在 SciPy 堆栈中，这主要是通过`SymPy`库实现的（请参见 [http： //docs.sympy.org/latest/index.html](http://docs.sympy.org/latest/index.html) ）。 符号微分和后验评估是替代非常基本功能的数字微分的一个不错的选择。 但是，通常，此方法会导致代码过于复杂和效率低下。 尽管可能出现错误，但纯数字微分的速度还是首选的。
*   Automatic differentiation is another set of techniques to numerically evaluate the derivative of a function. It is not based upon any approximation schema. This is without a doubt the most powerful option in the context of high derivatives of multivariate functions.

    ### 注意

    在 SciPy 堆栈中，这是通过不同的不相关库执行的。 一些最可靠的是`Theano`（ [http://deeplearning.net/software/theano/](http://deeplearning.net/software/theano/) ）或`FuncDesigner`（ [http://www.openopt.org/FuncDesigner](http://www.openopt.org/FuncDesigner) ）。 对于这些技术的全面描述和分析，可以在 [http://alexey.radul.name/ideas/2013/introduction-to-automatic-differentiation/](http://alexey.radul.name/ideas/2013/introduction-to-automatic-differentiation/) 中找到很好的资源。

## 数值微分

数值微分的最基本方案是使用节点间距均匀的中心差分公式执行的。 为了保持对称性，需要奇数个节点以保证较小的舍入误差。 此简单算法的实现在模块`scipy.misc`中可用。

### 提示

有关模块`scipy.misc`以及其基本例程的枚举的信息，请参考位于[的在线文档，网址为 http://docs.scipy.org/doc/scipy-0.13.0/reference/misc.html](http://docs.scipy.org/doc/scipy-0.13.0/reference/misc.html) 。

为了近似多项式函数的一阶和二阶导数，例如， *x = 1* 时， *f（x）= x* *<sup>5</sup>* 。 在距离`dx=1e-6`处有 15 个等距节点（以 *x = 1* 为中心）的情况下，我们可以发出以下命令：

```py
In [1]: import numpy as np
In [2]: from scipy.misc import derivative
In [3]: def f(x): return x**5
In [4]: derivative(f, 1.0, dx=1e-6, order=15)
Out[4]: 4.9999999997262723
In [5]: derivative(f, 1.0, dx=1e-6, order=15, n=2)
Out[5]: 19.998683310705456

```

由于实际值分别为`5`和`20`，因此有些准确，但仍然令人失望。

### 提示

该方法的另一个缺陷（至少相对于 SciPy 中编码的实现而言）是这样的事实，即结果依赖于可能很大的和，而这些和是不稳定的。 作为用户，我们可以通过使用 Shewchuk 算法修改`scipy.misc.derivative`源中的循环来改善问题。

## 象征差异

多项式的精确微分可以通过模块`numpy.polynomial`实现：

```py
In [6]: p = np.poly1d([1,0,0,0,0,0]); \
 ...: print p
 5
1 x
In [7]: np.polyder(p,1)(1.0)             In [7]: p.deriv()(1.0)
Out[7]: 5.0                              Out[7]: 5.0 
In [8]: np.polyder(p,2)(1.0)             In [8]: p.deriv(2)(1.0)
Out[8]: 20.0                             Out[8]: 20.0

```

符号差异是获得精确结果的另一种方法：

```py
In [9]: from sympy import diff, symbols
In [10]: x = symbols('x', real=True)
In [11]: diff(x**5, x)                   In [12]: diff(x**5, x, x)
Out[11]: 5*x**4                          Out[12]: 20*x**3
In [13]: diff(x**5, x).subs(x, 1.0) 
Out[13]: 5.00000000000000
In [14]: diff(x**5, x, x).subs(x, 1.0)
Out[14]: 20.0000000000000

```

当我们区分比简单多项式更复杂的函数时，请注意略有改进（在符号表示和简化编码方面）。 例如，对于 *g（x）= e* *<sup>-x</sup>* *sinx* 在 *x = 1* 时：

```py
In [15]: def g(x): return np.exp(-x) * np.sin(x)
In [16]: derivative(g, 1.0, dx=1e-6, order=101)
Out[16]: -0.11079376536871781
In [17]: from sympy import sin as Sin, exp as Exp
In [18]: diff(Exp(-x) * Sin(x), x).subs(x, 1.0)
Out[18]: -0.110793765306699

```

相对于其数值或自动对应项，符号微分的一个巨大优势是可以极其轻松地计算偏导数。 让我们通过计算多元函数 *h（x，y，z）= e* *<sup>xyz</sup>* 在 *x = 1 处的四阶导数来说明这一点。* ， *y = 1* 和 *z = 2* ：

```py
In [19]: y, z = symbols('y z', real=True)
In [20]: diff(Exp(x * y * z), z, z, y, x).subs({x:1.0, y:1.0, z:2.0})
Out[20]: 133.003009780752

```

## 自动区分

第三种方法采用自动区分。 对于此示例，我们将使用库`FuncDesigner`：

```py
In [21]: from FuncDesigner import oovar, exp as EXP, sin as SIN
In [22]: X = oovar('X'); \
 ....: G = EXP(-X) * SIN(X)
In [23]: G.D({X: 1.0}, X)
Out[23]: -0.11079376530669924

```

结果显然比通过数值微分获得的结果更准确。 另外，不需要提供任何额外的参数。

# 积分

为了在合适的域上实现功能的确定积分，我们主要有两种方法-**数值积分**和**符号** **积分**。

数值积分是指通过正交过程近似定积分。 根据给出函数 *f（x）*的方式，积分的域，其奇异性的知识以及正交的选择，我们可以采用不同的方法来解决此问题：

*   对于单变量多项式，在每个有限区间以代数形式实现精确积分
*   对于在其范围内作为有限样本集合给出的函数：
    *   复合梯形法则
    *   辛普森梯形法则
    *   Romberg 整合方案
*   对于以有限间隔作为 Python 函数给出的通用单变量函数：
    *   固定阶高斯正交
    *   固定公差高斯正交
    *   通过应用 21 点，43 点和 87 点的高斯-克朗定律，实现简单的非自适应正交
    *   简单的自适应正交，在每个子间隔上进行细分和正交
*   基于每个子间隔内 21 点高斯-克朗德积分的盲全局自适应正交，并具有加速过程（彼得·韦恩的 epsilon 算法）：
    *   基于先前的全局自适应正交，但具有用户提供的奇异点/间断点位置
    *   自适应 Romberg 集成方案
*   对于作为无限制区间上的 Python 函数给出的单变量函数，存在全局自适应正交。 该过程将无限间隔转换为半开放间隔，并在每个子间隔内应用 15 点高斯-克朗德积分。
*   对于在类型 I 域上将作为 Python 函数给出的多元函数（将在稍后进行描述），通常使用在每个维度上迭代应用自适应单变量正交的方法。

在许多情况下，借助符号计算，甚至对于无界域也可以执行精确的积分。 在 SciPy 堆栈中，为此，我们为基本函数提供了 Risch 算法的实现，为非基本积分提供了 Meijer G 函数。 两种方法都位于`SymPy`库中。 不幸的是，这些符号过程不适用于所有功能。 而且，由于所生成代码的复杂性，通常，用这种方法获得的解决不像任何数值近似那样快。

## 符号整合

可以通过微积分的基本定理，通过模块`numpy.polynomial`，非常精确地计算有限域`[a,b]`上多项式函数的定积分。 例如，要计算区间`[-1,1]`上多项式 *p（x）= x* *5* 的积分，我们可以发出：

```py
In [1]: import numpy as np
In [2]: p = np.poly1d([1,0,0,0,0,0]); \
 ...: print p; \
 ...: print p.integ()
 5
1 x
 6
0.1667 x
In [3]: p.integ()(1.0) - p.integ()(-1.0)
Out[3]: 0.0

```

通常，难以获得泛型函数的确定积分的精确值，并且计算效率低下。 在某些情况下，借助于 Risch 算法（对于基本函数）和 Meijer G 函数（对于非基本积分），可以通过符号积分来实现。 可以使用集成在库`SymPy`中的通用例程来调用这两种方法。 该例程足够聪明，可以根据源函数来决定使用哪种算法。

让我们向您展示一些示例，这些示例从前一种情况的多项式的定积分开始：

```py
In [4]: from sympy import integrate, symbols
In [5]: x, y = symbols('x y', real=True)
In [6]: integrate(x**5, x)
Out[6]: x**6/6
In [7]: integrate(x**5, (x, -1, 1))
Out[7]: 0

```

让我们尝试一些更复杂的事情。 函数 *f（x）= e* *<sup>-x</sup>* *sinx* 的定积分在区间`[0,1]`上：

```py
In [8]: from sympy import N, exp as Exp, sin as Sin
In [9]: integrate(Exp(-x) * Sin(x), x)
Out[9]: -exp(-x)*sin(x)/2 - exp(-x)*cos(x)/2
In [10]: integrate(Exp(-x) * Sin(x), (x, 0, 1))
Out[10]: -exp(-1)*sin(1)/2 - exp(-1)*cos(1)/2 + 1/2
In [11]: N(_)
Out[11]: 0.245837007000237

```

符号集成在工作时会以正确的方式对待奇点：

```py
In [12]: integrate(Sin(x) / x, x)
Out[12]: Si(x)
In [13]: integrate(Sin(x) / x, (x, 0, 1))
Out[13]: Si(1)
In [14]: N(_)
Out[14]: 0.946083070367183
In [15]: integrate(x**(1/x), (x, 0, 1))
Out[15]: 1/2

```

也可以在无界域上进行集成：

```py
In [16]: from sympy import oo
In [17]: integrate(Exp(-x**2), (x,0,+oo))
Out[17]: sqrt(pi)/2

```

甚至可以执行多变量集成：

```py
In [18]: integrate(Exp(-x**2-y**2), (x, -oo, +oo), (y, -oo, +oo))
Out[18]: pi

```

但是，我们需要特别强调这一点-符号集成在简单情况下效率不高（并且可能不起作用！），如以下示例所示：

```py
In [19]: integrate(Sin(x)**Sin(x), x)
Integral(sin(x)**sin(x), x)
In [20]: integrate(Sin(x)**Sin(x), (x, 0, 1))
Integral(sin(x)**sin(x), (x, 0, 1))

```

即使适用于简单情况，它也会生成复杂的代码，并且可能会使用过多的计算资源。

## 数值积分

解决这些问题的最佳方法是在数值积分的帮助下获得良好的近似值。 根据功能的类型和集成域，有不同的技术。 让我们详细研究它们。

### 在有限间隔上没有奇异的函数

数值积分中的基本问题是在有限间隔`[a,b]`上近似任何函数 *f（x）*的定积分。 通常，如果函数 *f（x）*不具有奇异性或不连续性，我们可以通过将不同的插值与分段多项式进行积分来获得简单的正交公式（因为对它们进行了精确评估）：

*   通过集成分段线性插值器（每两个连续的节点）来实现复合梯形法则
*   辛普森法则是通过集成分段多项式插值器来实现的，其中每两个连续的子间隔我们拟合一个抛物线
*   在前一种情况下，如果我们进一步施加 Hermite 插值，则可以获得复合 Simpson 规则

我们分别通过例程`cumtrapz`和`simps`在模块`scipy.integrate`中提供了用于复合梯形规则和复合 Simpson 规则的高效算法。 让我们向您展示如何将这些简单的正交公式用于多项式示例：

```py
In [21]: from scipy.integrate import cumtrapz, simps
In [22]: def f(x): return x**5
In [23]: nodes = np.linspace(-1, 1, 100)
In [24]: simps(f(nodes), nodes)
Out[24]: -1.3877787807814457e-17
In [25]: cumtrapz(f(nodes), nodes)
Out[25]:
array([ -1.92221161e-02,  -3.65619927e-02,  -5.21700680e-02,
 -6.61875756e-02,  -7.87469280e-02,  -8.99720915e-02,
 -9.99789539e-02,  -1.08875683e-01,  -1.16763077e-01,
 -1.23734908e-01,  -1.29878257e-01,  -1.35273836e-01,
 -1.39996314e-01,  -1.44114617e-01,  -1.47692240e-01,
 -1.50787532e-01,  -1.53453988e-01,  -1.55740523e-01,
 -1.57691741e-01,  -1.59348197e-01,  -1.60746651e-01,
 -1.61920310e-01,  -1.62899066e-01,  -1.63709727e-01,
 -1.64376231e-01,  -1.64919865e-01,  -1.65359463e-01,
 -1.65711607e-01,  -1.65990811e-01,  -1.66209700e-01,
 -1.66379187e-01,  -1.66508627e-01,  -1.66605982e-01,
 -1.66677959e-01,  -1.66730153e-01,  -1.66767180e-01,
 -1.66792794e-01,  -1.66810003e-01,  -1.66821177e-01,
 -1.66828145e-01,  -1.66832283e-01,  -1.66834598e-01,
 -1.66835799e-01,  -1.66836364e-01,  -1.66836598e-01,
 -1.66836678e-01,  -1.66836700e-01,  -1.66836703e-01,
 -1.66836703e-01,  -1.66836703e-01,  -1.66836703e-01,
 -1.66836700e-01,  -1.66836678e-01,  -1.66836598e-01,
 -1.66836364e-01,  -1.66835799e-01,  -1.66834598e-01,
 -1.66832283e-01,  -1.66828145e-01,  -1.66821177e-01,
 -1.66810003e-01,  -1.66792794e-01,  -1.66767180e-01,
 -1.66730153e-01,  -1.66677959e-01,  -1.66605982e-01,
 -1.66508627e-01,  -1.66379187e-01,  -1.66209700e-01,
 -1.65990811e-01,  -1.65711607e-01,  -1.65359463e-01,
 -1.64919865e-01,  -1.64376231e-01,  -1.63709727e-01,
 -1.62899066e-01,  -1.61920310e-01,  -1.60746651e-01,
 -1.59348197e-01,  -1.57691741e-01,  -1.55740523e-01,
 -1.53453988e-01,  -1.50787532e-01,  -1.47692240e-01,
 -1.44114617e-01,  -1.39996314e-01,  -1.35273836e-01,
 -1.29878257e-01,  -1.23734908e-01,  -1.16763077e-01,
 -1.08875683e-01,  -9.99789539e-02,  -8.99720915e-02,
 -7.87469280e-02,  -6.61875756e-02,  -5.21700680e-02,
 -3.65619927e-02,  -1.92221161e-02,  -1.73472348e-17])

```

### 提示

例程`cumtrapz`计算指定子间隔内的累积积分。 因此，输出的最后一项是我们寻求的正交值。 当然，我们可以通过简单地访问该条目来仅报告所需的积分：

```py
In [26]: cumtrapz(f(nodes), nodes)[-1]
Out[26]: -1.7347234759768071e-17

```

这两种算法的实现未明确计算插值器。 最终公式是这里的目标，并且它在 SciPy 中的编码方式是通过 Newton-Cotes 求积法。

执行 Newton-Cotes 的例程是隐藏的（在某种意义上说，它们未在 SciPy 的官方页面上的教程或文档中进行报告），​​并且仅供`cumtrapz`或`simps`内部使用。 它们仅提供相应的系数，这些系数会乘以节点处的功能评估。

但是，在正确的情况下，牛顿-科特斯正交公式通常通常非常准确。 在许多情况下，它们可以用于计算更好的近似值，而不必遵守梯形或辛普森规则。

让我们展示一下它如何在我们的运行示例中工作，现在在间隔`[-1,1]`中只有四个等距节点：

```py
In [27]: from scipy.integrate import newton_cotes
In [28]: coefficients, abs_error = newton_cotes(3, equal=True); \
 ....: nodes = np.linspace(-1, 1, 4); \
 ....: print coefficients
[ 0.375  1.125  1.125  0.375]
In [29]: integral = (coefficients * f(nodes)).sum(); \
 ....: print integral
0.0
In [30]: from math import fsum
In [31]: integral = fsum(coefficients * f(nodes)); \
 ....: print integral
-7.8062556419e-18

```

如果我们选择的节点之间的间距相等，则在特殊情况下，如果子间隔的数量为 2 的幂，则梯形规则将得到改进。 在这种情况下，我们可以使用 Romberg 规则-使用 Richardson 外推法的一种改进。 我们可以在同一模块中使用例程`romb`访问它。

让我们将结果与正在运行的示例进行比较，这次使用间隔`[-1,1]`中的大小为`1/32`的 64 个子间隔：

```py
In [32]: from scipy.integrate import romb
In [33]: nodes = np.linspace(-1, 1, 65)
In [34]: romb(f(nodes), dx=1./32)
0.0

```

我们可以选择报告表，该表显示了从给定节点进行的 Richardson 外推：

```py
In [35]: romb(f(nodes), dx=1./32, show=True)
 Richardson Extrapolation Table for Romberg Integration 
====================================================================
 0.00000
 0.00000  0.00000
 0.00000  0.00000  0.00000
 0.00000  0.00000  0.00000  0.00000
 0.00000  0.00000  0.00000  0.00000  0.00000
 0.00000  0.00000  0.00000  0.00000  0.00000  0.00000
====================================================================
Out[35]: 0.0

```

对于节点的选择，我们可能没有任何偏好，但是对于我们的数值积分方案，使用 Romberg 规则仍然有我们的想法。 在这种情况下，我们可以使用例程`romberg`，对于该例程，我们只需要提供函数的表达式和积分的限制即可。 （可选）我们可以提供误差的绝对或相对公差（默认均设置为`1.48e-8`）：

```py
In [36]: from scipy.integrate import romberg
In [37]: romberg(f, -1, 1, show=True)
Romberg integration of <function vfunc at 0x10ffa8c08> from [-1, 1]
 Steps  StepSize   Results
 1  2.000000  0.000000
 2  1.000000  0.000000  0.000000
The final result is 0.0 after 3 function evaluations.
Out[37]: 0.0

```

另一种可能性是使用高斯正交公式。 由于通过在内部计算节点的最佳可能选择来获得近似值的精度，因此这些功能更为强大。 模块`scipy.integrate`中有两个基本例程执行该算法的实现：`quadrature`（如果我们要指定容差），或`fixed_quad`（如果我们要指定节点数（而不是它们的位置！））。 ：

```py
In [38]: from scipy.integrate import quadrature, fixed_quad
In [39]: value, absolute_error = quadrature(f, -1, 1, tol=1.49e-8); \
 ....: print value
0.0
In [40]: value, absolute_error = fixed_quad(f, -1, 1, n=4); \
 ....: print value                                  # four nodes
-9.45424294407e-16

```

通过模块`scipy.integrate`中的函数`quad`获得使用自适应方案执行高斯正交的更高级的方法。 此函数是 Fortran 库`QUADPACK`中例程`QAGS`的包装。 该算法将积分域划分为几个子间隔，并在每个子间隔上执行 21 点高斯-克朗罗德正交规则。 借助 Peter Wynn 的 epsilon 算法，可以实现进一步的加速。

### 提示

有关`QAGS`以及`QUADPACK`库中其他例程的更多信息，请参考 netlib 存储库： [http://www.netlib.org/quadpack/](http://www.netlib.org/quadpack/) 。

让我们将其与正在运行的示例进行比较：

```py
In [41]: from scipy.integrate import quad
In [42]: value, absolute_error = quad(f, -1, 1); \
 ....: print value
0.0

```

我们可以通过将可选参数`full_output`设置为`True`来获得实现细节。 这为我们提供了额外的 Python 字典，其中包含有用的信息：

```py
In [43]: value, abs_error, info = quad(f, -1, 1, full_output=True)
In [44]: info.keys()
Out[44]: ['rlist', 'last', 'elist', 'iord', 'alist', 'blist',
 'neval']
In [45]: print "{0} function evaluations".format(info['neval'])
21 function evaluations
In [46]: print "Used {0} subintervals".format(info['last'])
Used 1 subintervals

```

为了完全理解`info`的所有不同输出，我们需要了解计算高斯正交的基础算法。 这些特殊例程使用基于 Chebyshev 矩的 Clensaw-Curtis 方法。

在前面的示例中，默认情况下，该代码尝试使用 50 个 Chebyshev 矩。 由于被整数的简单性，并且由于仅需要一个子间隔，因此仅需使用这些矩之一。 当我们从字典信息中报告 50 个条目的一维输出`rlist`，`elist`，`alist`和`blist`时，我们可以不理会每个条目的最后 49 个条目提供的信息：

```py
In [47]: np.set_printoptions(precision=2, suppress=True)
In [48]: print info['rlist']  # integral approx on subintervals
[  0.00e+000   2.32e+077   6.93e-310   0.00e+000   0.00e+000 
 0.00e+000   0.00e+000   0.00e+000   0.00e+000   0.00e+000 
 0.00e+000   0.00e+000   0.00e+000   0.00e+000   0.00e+000 
 0.00e+000   0.00e+000   0.00e+000   0.00e+000   0.00e+000 
 0.00e+000   6.45e-314   2.19e-314   6.93e-310   0.00e+000 
 0.00e+000   0.00e+000   0.00e+000   0.00e+000   0.00e+000
 0.00e+000   0.00e+000   0.00e+000   0.00e+000   0.00e+000 
 0.00e+000   0.00e+000   0.00e+000   0.00e+000   0.00e+000 
 0.00e+000   0.00e+000  -1.48e-224   2.19e-314   6.93e-310 
 0.00e+000   0.00e+000   0.00e+000   0.00e+000   0.00e+000]
In [49]: print info['elist']  # abs error estimates on subintervals
[  3.70e-015   2.32e+077   3.41e-322   0.00e+000   0.00e+000 
 0.00e+000   0.00e+000   0.00e+000   0.00e+000   0.00e+000 
 0.00e+000   0.00e+000   0.00e+000   0.00e+000   0.00e+000 
 0.00e+000   0.00e+000   0.00e+000   0.00e+000   7.30e+245 
 2.19e-314   6.93e-310   0.00e+000   0.00e+000   0.00e+000 
 4.74e+246   2.20e-314   6.93e-310   0.00e+000   0.00e+000
 0.00e+000   0.00e+000   0.00e+000   0.00e+000  -9.52e+207 
 2.19e-314   6.93e-310   0.00e+000   0.00e+000   0.00e+000 
 0.00e+000   0.00e+000   0.00e+000   0.00e+000   0.00e+000 
 0.00e+000   2.00e+000   2.00e+000   2.27e-322   1.05e-319]
In [50]: print info['alist']   # subintervals left end points
[-1\.  2\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\. 
 0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\. 
 0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\. 
 0\.  0\.  0\.  0\.  0.]
In [51]: print info['blist']   # subintervals right end pts
[ 1\.  2\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\. 
 0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\. 
 0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\.  0\. 
 0\.  0\.  0\.  2\. -0.]

```

可以施加不同数量的切比雪夫矩。 我们使用可选参数`maxp1`来执行此操作，该参数对该数字强加一个上限（而不是固定该数字，以获得最佳结果）。

即使 *f（x）*光滑，形式为 *f（x）cos（wx）*或 *f（x）sin（wx）*的振荡积分， 尤其棘手。 积分器`quad`可以通过调用`QUADPACK`中的例程`QAWO`处理这些积分。 我们可以通过使用`wvar=w`指定参数`weight='cos'`或`weight='sin'`来采用此方法。

例如，对于区间`[-10,10]`上 *g（x）= sin（x）e* *<sup>x</sup>* 的积分，我们将此方法与基本 `quad`。 我们可以执行以下操作：

```py
In [52]: def f(x): return np.sin(x) * np.exp(x)
In [53]: g = np.exp
In [54]: quad(g, -10, 10, weight='sin', wvar=1)
Out[54]: (3249.4589405744427, 5.365398805302381e-08)
In [55]: quad(f, -10, 10)
Out[55]: (3249.458940574436, 1.1767634585879705e-05)

```

注意绝对误差的显着增加。

### 提示

有关本节中探讨的所有正交公式的详细信息和理论，请参考第 3 章，**的数值微分和积分* Walter Gautchi 的数值分析[* 。

### 有界域上具有奇异性的函数

积分的第二种情况是具有奇异性的函数在有限区间`[a,b]`上的确定积分。 我们考虑两种情况：加权函数和具有奇点的泛型函数。

#### 加权函数

加权函数可以实现为 *f（x）w（x）*类型的乘积，用于某些平滑函数 *f（x）*和非负加权函数 *w（ x）*包含奇点。 *cos（πx/ 2）/* *√* *x* 给出了一个说明性示例。 我们可以将这种情况视为 *cos（πx/ 2）*与 *w（x）= 1 /* *√* *x* 的乘积。 权重在 *x = 0* 处呈现单个奇点，否则平滑。

处理这些积分的常用方法是借助加权高斯正交公式。 例如，要执行 f *（x）/（x-c）*形式的函数的主值积分，我们向`quad`发出参数`weight='cauchy'`和`wvar=c`。 这从`QUADPACK`调用例程`QAWC`。

让我们在间隔`[-1,1]`上试验 *g（x）= sin（x）/ x* 的菲涅耳型正弦积分，并将其与`romberg`进行比较：

```py
In [56]: value, abs_error = quad(f, -1, 1, weight='cauchy',wvar=0); \
 ....: print value
1.89216614073
In [57]: romberg(g, -1, 1)
Out[57]: 2.35040238729

```

在具有权重的函数积分的情况下*（xa）* *<sup>α</sup>* *（bx）* *<sup>β</sup>* ，其中`a`和`b`是整合域的端点，并且`alpha`和`beta`都大于`-1`，我们向`quad`发出了参数`weight='alg'`和`wvar=(alpha, beta)`。 这从`QUADPACK`调用例程`QAWS`。

让我们用 *g（x）= cos（πx/ 2）/* *√* *x* 的菲涅耳型余弦积分进行实验，并进行比较 与`quadrature`一起使用：

```py
In [58]: def f(x): return np.cos(np.pi * x * 0.5)
In [59]: def g(x): return np.cos(np.pi * x * 0.5) / np.sqrt(x)
In [60]: value, abs_error = quad(f, 0, 1, weight='alg', \
 wvar=(-0.5,0)); \
 ....: print value
1.55978680075
In [61]: quadrature(g, 0, 1)
quadrature.py:178: AccuracyWarning: maxiter (50) exceeded. Latest
difference = 3.483182e-04
 AccuracyWarning)
Out[61]: (1.5425452942607543, 0.00034831815190772275)

```

如果权重的形式为 *w（x）=（xa）* *<sup>α</sup>* *（bx）* *<sup>β[</sup>* *ln（xa）*， *w（x）=（xa）* *<sup>α</sup>* *（bx）* *<sup>β</sup>* *ln（bx）*或 *w（x）=（xa）* *<sup>α</sup>* *（bx）* *<sup>β</sup>* *ln（xa）ln（bx）*，我们用以下参数发布`quad`： 分别为`weight='alg-loga'`或`weight='alg-logb'`或`weight='alg-log'`，并分别为`wvar=(alpha, beta)`。 例如，对于间隔`[0,1]`上的函数 *g（x）= x* *<sup>2</sup>* *ln（x）* 发出以下命令：

```py
In [62]: def f(x): return x**2
In [63]: def g(x): return x**2 * np.log(x)
In [64]: value, abs_error = quad(f, 0, 1, weight='alg-loga', \
 wvar=(0,0)); \
 ....: print value
-0.111111111111

```

该积分的实际值为`-1/9`。

#### 具有奇异性的一般功能

通常，我们可能会以奇异的形式处理函数，这些奇数不符合上一节中指出的漂亮形式 *f（x）w（x）*。 在这种情况下，如果我们知道奇点的位置，则可以向带有可选参数点的积分器四边形表示。 积分器`quad`调用`QUADPACK`中的例程`QAGP`。 例如，对于函数 *g（x）= floor（x）ln（x）*，它在每个整数上观察到一个奇数，以对区间`[1,8]`进行积分，我们可以发出以下命令：

```py
In [65]: def g(x): return np.floor(x) * np.log(x)
In [66]: quad(g, 1, 8, points=np.arange(8)+1)
Out[66]: (45.802300241541005, 5.085076830895319e-13)

```

将此与简单的四边形计算进行比较，而无需指出任何奇点，如下面的代码行所示：

```py
In [67]: quad(g, 1, 8)
quadpack.py:295: UserWarning: The maximum number of subdivisions (50)
has been achieved.
 If increasing the limit yields no improvement it is advised to
analyze the integrand in order to determine the difficulties.  If the
position of a local difficulty can be determined (singularity,
discontinuity) one will probably gain from splitting up the interval
and calling the integrator on the subranges.  Perhaps a special-
purpose integrator should be used.
 warnings.warn(msg)
Out[67]: (45.80231242134967, 8.09803045669355e-05)

```

### 在无界域上集成

通过调用`QUADPACK`中的例程`QAGI`，通用积分器`quad`还能够使用自适应正交公式在无界域上计算定积分。 此过程不适用于'`cauchy`'或任何'`alg`'型砝码选项。

通常，如果要集成的函数不具有奇异性，则近似值是可靠的。 奇异性的存在给出不可靠的积分，如以下示例所示：

```py
In [68]: def f(x): return 2 * np.exp(-x**2) / np.sqrt(np.pi)
In [69]: value, absolute_error = quad(f, 0, np.inf); \
 ....: print value
1.0
In [70]: def f(x): return np.sin(x)/x
In [71]: integrate(Sin(x)/x, (x, 0, oo))
Out[71]: pi/2
In [72]: value, absolute_error = quad(f, 0, np.inf); \
 ....: print value                                      # ouch!
2.24786796347

```

对于无界域中的振荡积分，除了发出带有参数`weight='cos'`或`weight='sin'`以及相应的`wvar`参数的`quad`之外，我们还可以为内部使用的循环数设置上限。 为此，我们将可选参数`limlst`设置为所需的界限。 通常最好将其设置为大于三的值。 例如，对于 *[1，∞]* 上`sinc`函数的傅里叶积分，我们可以发出以下命令：

```py
In [73]: def f(x): return 1./x
In [74]: quad(f, 1, np.inf weight='sin', wvar=1, limlst=5)
quadpack.py:295: UserWarning: The maximum number of cycles allowed
has been achieved., e.e. of subintervals (a+(k-1)c, a+kc) where c =
(2*int(abs(omega)+1))*pi/abs(omega), for k = 1, 2, ..., lst.  One can
allow more cycles by increasing the value of limlst.  Look at
info['ierlst'] with full_output=1.
 warnings.warn(msg)
Out[74]: (0.636293340511029, 1.3041427865109276)
In [75]: quad(f, 1, np.inf, weight='sin', wvar=1, limlst=50)
Out[75]: (0.6247132564795975, 1.4220476353655983e-08)

```

## 数值多元积分

通过应用自适应高斯正交规则，还可以在不同的域上执行多元数值积分。 为此，在模块`scipy.integrate`中，我们必须执行例程`dblquad`（双积分），`tplquad`（三重积分）和`nquad`（多个变量的积分）。

这些例程只能在 I 型区域上计算定积分：

*   在两个维度中，类型 I 域可以以 *{（x，y）的形式写：a < x < b，f（x）< y < h（x）}* 代表两个数字`a`和`b`和两个单变量函数 *f（x）*和 *h（x）*。
*   在三个维度上，类型 I 区域可以以 *{（x，y，z）的形式写：a < x < b，f（x）< y < h（x ），q（x，y）< z < r（x，y）}* 对于数字`a`，`b`，单变量函数 *f（x）*， *h（x）*和二元函数 *q（x，y），r（x，y）*。
*   在三个以上的维度中，I 型区域可以按与其双重和三重对应区域相似的方式顺序写入。 第一个变量以两个数字为界。 第二个变量受第一个变量的两个单变量函数限制。 第三个变量受两个第一个变量的两个双变量函数限制，依此类推。

让我们对`In [18]`行中的示例函数进行数值积分。 请注意在要集成的函数的定义中必须引入不同变量的顺序：

```py
In [76]: def f(x, y): return np.exp(-x**2 - y**2)
In [77]: from scipy.integrate import dblquad
In [78]: dblquad(f, 0, np.inf, lambda x:0, lambda x:np.inf)
Out[78]: (0.785398163397,  6.29467149642e-09)

```

# 摘要

在本章中，我们已经掌握了计算功能差异和集成的所有不同方法。 我们了解到`scipy`库具有非常健壮的例程，可以用数值方式计算所有这些操作的近似值（必要时包装有效的 Fortran 库）。 我们还了解到，可以访问 SciPy 堆栈中的其他库以符号或自动方式执行操作。

在下一章中，我们将探讨在非线性函数的背景下求解方程或方程组的理论和方法，以及为优化目的而计算极值。