# 八、通过测试确保质量

一些程序员仅在生产中进行测试。 如果您不是其中之一，那么您可能熟悉单元测试的概念。 单元测试是程序员编写的用于测试其代码的自动测试。 例如，这些测试可以单独测试函数或函数的一部分。 每个测试仅覆盖一小部分代码。 这样做的好处是提高了对代码质量，可重复测试的信心，并附带了更清晰的代码。

Python 对单元测试有很好的支持。 此外，NumPy 将`numpy.testing`包添加到 NumPy 代码单元测试的包中。

**测试驱动的开发**（**TDD**）是最重要的事情之一发生在软件开发中。 TDD 将集中在自动化单元测试上。 目标是尽可能自动地测试代码。 下次更改代码时，我们可以运行测试并捕获潜在的回归。 换句话说，任何已经存在的函数仍然可以使用。

本章中的主题包括：

*   单元测试
*   断言
*   浮点精度

## 断言函数

单元测试通常使用函数，这些函数断言某些内容是测试的一部分。 在进行数值计算时，通常存在一个基本问题，即试图比较几乎相等的浮点数。 对于整数，比较是微不足道的操作，但对于浮点数则不是，因为计算机的表示不准确。 NumPy `testing`包具有许多工具函数，这些函数可以测试先决条件是否成立，同时考虑到浮点比较的问题。 下表显示了不同的工具函数：

| 函数 | 描述 |
| --- | --- |
| `assert_almost_equal()` | 如果两个数字不等于指定的精度，则此函数引发异常 |
| `assert_approx_equal()` | 如果两个数字在一定意义上不相等，则此函数引发异常 |
| `assert_array_almost_equal()` | 如果两个数组的指定精度不相等，此函数将引发异常 |
| `assert_array_equal()` | 如果两个数组不相等，此函数将引发异常。 |
| `assert_array_less()` | 如果两个数组的形状不同，并且第一个数组的元素严格小于第二个数组的元素，则此函数引发异常 |
| `assert_equal()` | 如果两个对象不相等，则此函数引发异常 |
| `assert_raises()` | 如果使用定义的参数调用的可调用对象未引发指定的异常，则此函数失败 |
| `assert_warns()` | 如果未抛出指定的警告，则此函数失败 |
| `assert_string_equal()` | 此函数断言两个字符串相等 |
| `assert_allclose()` | 如果两个对象不等于期望的公差，则此函数引发断言 |

## 实战时间 – 断言几乎相等

假设您有两个几乎相等的数字。 让我们使用`assert_almost_equal()`函数检查它们是否相等：

1.  以较低精度调用函数（最多 7 个小数位）：

    ```py
    print("Decimal 6", np.testing.assert_almost_equal(0.123456789, 0.123456780, decimal=7))
    ```

    请注意，不会引发异常，如以下结果所示：

    ```py
    Decimal 6 None

    ```

2.  以更高的精度调用该函数（最多 8 个小数位）：

    ```py
    print("Decimal 7", np.testing.assert_almost_equal(0.123456789, 0.123456780, decimal=8))
    ```

    结果如下：

    ```py
    Decimal 7
    Traceback (most recent call last):
     …
     raise AssertionError(msg)
    AssertionError:
    Arrays are not almost equal
     ACTUAL: 0.123456789
     DESIRED: 0.12345678

    ```

### 刚刚发生了什么？

我们使用了 NumPy `testing`包中的`assert_almost_equal()`函数来检查`0.123456789`和`0.123456780`对于不同的十进制精度是否相等。

### 小测验 - 指定小数精度

Q1. `assert_almost_equal()`函数的哪个参数指定小数精度？

1.  `decimal`
2.  `precision`
3.  `tolerance`
4.  `significant`

## 近似相等的数组

如果两个数字在一定数量的有效数字下不相等，则`assert_approx_equal()`函数会引发异常。 该函数引发由以下情况触发的异常：

```py
abs(actual - expected) >= 10**-(significant - 1)
```

## 实战时间 – 断言近似相等

让我们从上一个“实战”部分中选取数字，在它们上应用`assert_approx_equal()`函数：

1.  以低重要性调用函数：

    ```py
    print("Significance 8", np.testing.assert_approx_equal (0.123456789, 0.123456780,significant=8))
    ```

    The result is as follows:

    ```py
    Significance 8 None

    ```

2.  以高重要性调用函数：

    ```py
    print("Significance 9", np.testing.assert_approx_equal (0.123456789, 0.123456780, significant=9))
    ```

    该函数引发一个`AssertionError`：

    ```py
    Significance 9
    Traceback (most recent call last):
     ...
     raise AssertionError(msg)
    AssertionError:
    Items are not equal to 9 significant digits:
     ACTUAL: 0.123456789
     DESIRED: 0.12345678

    ```

### 刚刚发生了什么？

我们使用了 NumPy `testing`包中的`assert_approx_equal()`函数来检查`0.123456789`和`0.123456780`对于不同的十进制精度是否相等。

## 几乎相等的数组

如果两个数组在指定的精度下不相等，则`assert_array_almost_equal()`函数会引发异常。 该函数检查两个数组的形状是否相同。 然后，将数组的值与以下元素进行逐元素比较：

```py
|expected - actual| < 0.5 10-decimal
```

## 实战时间 – 断言数组几乎相等

通过向每个数组添加`0`，用上一个“实战时间”部分的值构成数组：

1.  以较低的精度调用该函数：

    ```py
    print("Decimal 8", np.testing.assert_array_almost_equal([0, 0.123456789], [0, 0.123456780], decimal=8))
    ```

    The result is as follows:

    ```py
    Decimal 8 None

    ```

2.  以较高的精度调用该函数：

    ```py
    print("Decimal 9", np.testing.assert_array_almost_equal([0, 0.123456789], [0, 0.123456780], decimal=9))
    ```

    测试产生一个`AssertionError`：

    ```py
    Decimal 9
    Traceback (most recent call last):
     …
     assert_array_compare
     raise AssertionError(msg)
    AssertionError:
    Arrays are not almost equal

    (mismatch 50.0%)
     x: array([ 0\.        ,  0.12345679])
     y: array([ 0\.        ,  0.12345678])

    ```

### 刚刚发生了什么？

我们将两个数组与 NumPy `array_almost_equal()`函数进行了比较。

### 勇往直前 – 比较不同形状的数组

使用 NumPy `array_almost_equal()`函数比较具有不同形状的两个数组。

## 相等的数组

如果两个数组不相等，`assert_array_equal()`函数将引发异常。 数组的形状必须相等，并且每个数组的元素必须相等。 数组中允许使用 NaN。 或者，可以将数组与`array_allclose()`函数进行比较。 此函数的参数为**绝对公差**（`atol`）和**相对公差**（`rtol`）。 对于两个数组`a`和`b`，这些参数满足以下方程式：

```py
|a - b| <= (atol + rtol * |b|)
```

## 实战时间 – 比较数组

让我们将两个数组与刚才提到的函数进行比较。 我们将重复使用先前“实战”中的数组，并将它们加上 NaN：

1.  调用`array_allclose()`函数：

    ```py
    print("Pass", np.testing.assert_allclose([0, 0.123456789, np.nan], [0, 0.123456780, np.nan], rtol=1e-7, atol=0))
    ```

    The result is as follows:

    ```py
    Pass None

    ```

2.  调用`array_equal()`函数：

    ```py
    print("Fail", np.testing.assert_array_equal([0, 0.123456789, np.nan], [0, 0.123456780, np.nan]))
    ```

    测试失败，并显示`AssertionError`：

    ```py
    Fail
    Traceback (most recent call last):
     …
    assert_array_compare
     raise AssertionError(msg)
    AssertionError:
    Arrays are not equal

    (mismatch 50.0%)
     x: array([ 0\.        ,  0.12345679,         nan])
     y: array([ 0\.        ,  0.12345678,         nan])

    ```

### 刚刚发生了什么？

我们将两个数组与`array_allclose()`函数和`array_equal()`函数进行了比较。

## 排序数组

如果两个数组不具有相同形状的，并且第一个数组的元素严格小于第二个数组的元素，则`assert_array_less()`函数会引发异常。

## 实战时间 – 检查数组顺序

让我们检查一个数组是否严格大于另一个数组：

1.  用两个严格排序的数组调用`assert_array_less()`函数：

    ```py
    print("Pass", np.testing.assert_array_less([0, 0.123456789, np.nan], [1, 0.23456780, np.nan]))
    ```

    The result is as follows:

    ```py
    Pass None

    ```

2.  调用`assert_array_less()`函数：

    ```py
    print("Fail", np.testing.assert_array_less([0, 0.123456789, np.nan], [0, 0.123456780, np.nan]))
    ```

    该测试引发一个异常：

    ```py
    Fail
    Traceback (most recent call last):
     ...
     raise AssertionError(msg)
    AssertionError:
    Arrays are not less-ordered

    (mismatch 100.0%)
     x: array([ 0\.        ,  0.12345679,         nan])
     y: array([ 0\.        ,  0.12345678,         nan])

    ```

### 刚刚发生了什么？

我们使用`assert_array_less()`函数检查了两个数组的顺序。

## 对象比较

如果两个对象不相等，则`assert_equal()`函数将引发异常。 对象不必是 NumPy 数组，它们也可以是列表，元组或字典。

## 实战时间 – 比较对象

假设您需要比较两个元组。 我们可以使用`assert_equal()`函数来做到这一点。

调用`assert_equal()`函数：

```py
print("Equal?", np.testing.assert_equal((1, 2), (1, 3)))
```

该调用引发错误，因为项目不相等：

```py
Equal?
Traceback (most recent call last):
 ...
 raise AssertionError(msg)
AssertionError:
Items are not equal:
item=1

 ACTUAL: 2
 DESIRED: 3

```

### 刚刚发生了什么？

我们将两个元组与`assert_equal()`函数进行了比较-由于元组彼此不相等，因此引发了一个例外。

## 字符串比较

`assert_string_equal()`函数断言两个字符串相等。 如果测试失败，该函数将引发异常，并显示字符串之间的差异。 字符串字符的大小写很重要。

## 实战时间 – 比较字符串

让我们比较一下字符串。 这两个字符串都是单词`NumPy`：

1.  调用`assert_string_equal()`函数将字符串与自身进行比较。 该测试当然应该通过：

    ```py
    print("Pass", np.testing.assert_string_equal("NumPy", "NumPy"))
    ```

    测试通过：

    ```py
    Pass None

    ```

2.  调用`assert_string_equal()`函数将一个字符串与另一个字母相同但大小写不同的字符串进行比较。 此测试应引发异常：

    ```py
    print("Fail", np.testing.assert_string_equal("NumPy", "Numpy"))
    ```

    测试引发错误：

    ```py
    Fail
    Traceback (most recent call last):
     …
     raise AssertionError(msg)
    AssertionError: Differences in strings:
    - NumPy?    ^
    + Numpy?    ^

    ```

### 刚刚发生了什么？

我们将两个字符串与`assert_string_equal()`函数进行了比较。 当外壳不匹配时，该测试引发了异常。

## 浮点比较

计算机中浮点数的表示形式不准确。 比较浮点数时，这会导致问题。 `assert_array_almost_equal_nulp()`和`assert_array_max_ulp()` NumPy 函数提供一致的浮点比较。浮点数的**最低精度的单位**（**ULP**），根据 IEEE 754 规范，是基本算术运算所需的半精度。 您可以将此与标尺进行比较。 公制标尺通常具有毫米的刻度，但超过该刻度则只能估计半毫米。

机器`ε`是浮点算术中最大的相对舍入误差。 机器`ε`等于 ULP 相对于 1。NumPy `finfo()`函数使我们能够确定机器`ε`。 Python 标准库还可以为您提供机器的`ε`值。 该值应与 NumPy 给出的值相同。

## 实战时间 – 使用`assert_array_almost_equal_nulp`来比较

让我们看到`assert_array_almost_equal_nulp()`函数的作用：

1.  使用`finfo()`函数确定机器`epsilon`：

    ```py
    eps = np.finfo(float).eps
    print("EPS", eps)
    ```

    `ε`将如下所示：

    ```py
    EPS 2.22044604925e-16

    ```

2.  使用`assert_almost_equal_nulp()`函数将`1.0`与`1 + epsilon`进行比较。 对`1 + 2 * epsilon`执行相同的操作：

    ```py
    print("1", np.testing.assert_array_almost_equal_nulp(1.0, 1.0 + eps))
    print("2", np.testing.assert_array_almost_equal_nulp(1.0, 1.0 + 2 * eps))
    ```

    The result is as follows:

    ```py
    1 None
    2
    Traceback (most recent call last):
     …
     assert_array_almost_equal_nulp
     raise AssertionError(msg)
    AssertionError: X and Y are not equal to 1 ULP (max is 2)

    ```

### 刚刚发生了什么？

我们通过`finfo()`函数确定了机器`ε`。 然后，我们将`1.0`与`1 + epsilon`与`assert_almost_equal_nulp()`函数进行了比较。 但是，该测试通过了，添加另一个`ε`导致异常。

## 更多使用 ULP 的浮点比较

`assert_array_max_ulp()`函数允许您指定允许的 ULP 数量的上限。 `maxulp`参数接受整数作为限制。 默认情况下，此参数的值为 1。

## 实战时间 – 使用最大值 2 的比较

让我们进行与先前“实战”部分相同的事情，但在必要时， 指定`maxulp`为`2`：

1.  使用`finfo()`函数确定机器`epsilon`：

    ```py
    eps = np.finfo(float).eps
    print("EPS", eps)
    ```

    The epsilon would be as follows:

    ```py
    EPS 2.22044604925e-16

    ```

2.  按照前面的“实战时间”部分中进行的比较，但是将`assert_array_max_ulp()`函数与相应的`maxulp`值一起使用：

    ```py
    print("1", np.testing.assert_array_max_ulp(1.0, 1.0 + eps))
    print("2", np.testing.assert_array_max_ulp(1.0, 1 + 2 * eps, maxulp=2))
    ```

    输出为  ，如下所示：

    ```py
    1 1.0
    2 2.0

    ```

### 刚刚发生了什么？

我们比较了与之前“实战”部分相同的值，但在第二次比较中指定了`2`的`maxulp`。 通过将`assert_array_max_ulp()`函数与适当的`maxulp`值一起使用，这些测试通过了 ULP 数量返回值。

## 单元测试

单元测试是自动化测试，它测试一小段代码，通常是函数或方法。 Python 具有用于单元测试的`PyUnit` API。 作为 NumPy 的用户，我们可以利用之前在操作中看到的`assert`函数。

## 实战时间 – 编写单元测试

我们将为一个简单的阶乘函数编写测试  。 测试将检查所谓的快乐路径和异常状况。

1.  首先编写阶乘函数：

    ```py
    import numpy as np
    import unittest

    def factorial(n):
       if n == 0:
          return 1

       if n < 0:
          raise ValueError, "Unexpected negative value"

       return np.arange(1, n+1).cumprod()
    ```

    该代码使用`arange()`和`cumprod()`函数创建数组并计算累积乘积，但是我们添加了一些边界条件检查。

2.  现在我们将编写单元测试。 让我们写一个包含单元测试的类。 它从标准测试 Pytho 的`unittest`模块扩展了`TestCase`类。 测试具有以下三个属性的阶乘函数的调用：

    *   正数，正确的方式
    *   边界条件 0
    *   负数，这将导致错误

        ```py
        class FactorialTest(unittest.TestCase):
           def test_factorial(self):
              #Test for the factorial of 3 that should pass.
              self.assertEqual(6, factorial(3)[-1])
              np.testing.assert_equal(np.array([1, 2, 6]), factorial(3))

           def test_zero(self):
              #Test for the factorial of 0 that should pass.
              self.assertEqual(1, factorial(0))

           def test_negative(self):
              #Test for the factorial of negative numbers that should fail.
              # It should throw a ValueError, but we expect IndexError
              self.assertRaises(IndexError, factorial(-10))
        ```

        如以下输出所示，我们将其中一项测试失败了：

        ```py
        $ python unit_test.py
        .E.
        ======================================================================
        ERROR: test_negative (__main__.FactorialTest)
        ----------------------------------------------------------------------
        Traceback (most recent call last):
         File "unit_test.py", line 26, in test_negative
         self.assertRaises(IndexError, factorial(-10))
         File "unit_test.py", line 9, in factorial
         raise ValueError, "Unexpected negative value"
        ValueError: Unexpected negative value

        ----------------------------------------------------------------------
        Ran 3 tests in 0.003s

        FAILED (errors=1)

        ```

### 刚刚发生了什么？

我们对阶乘函数代码进行了一些满意的路径测试。 我们让边界条件测试故意失败（请参阅`unit_test.py`）：

```py
import numpy as np
import unittest

def factorial(n):
   if n == 0:
      return 1

   if n < 0:
      raise ValueError, "Unexpected negative value"

   return np.arange(1, n+1).cumprod()

class FactorialTest(unittest.TestCase):
   def test_factorial(self):
      #Test for the factorial of 3 that should pass.
      self.assertEqual(6, factorial(3)[-1])
      np.testing.assert_equal(np.array([1, 2, 6]), factorial(3))

   def test_zero(self):
      #Test for the factorial of 0 that should pass.
      self.assertEqual(1, factorial(0))

   def test_negative(self):
      #Test for the factorial of negative numbers that should fail.
      # It should throw a ValueError, but we expect IndexError
      self.assertRaises(IndexError, factorial(-10))

if __name__ == '__main__':
    unittest.main()
```

## `Nose`测试装饰器

鼻子是嘴巴上方的器官，人类和动物用来呼吸和闻味。 它也是一个 Python 框架  ，使（单元）测试变得更加容易。 `Nose`可帮助您组织测试。 根据`nose`文档：

> “将收集与`testMatch`正则表达式（默认值：`(?:^|[b_.-])[Tt]est`）匹配的任何 python 源文件，目录或包作为测试。”

`Nose`大量使用装饰器。 Python 装饰器是[指示有关方法或函数的注释](http://thecodeship.com/patterns/guide-to-python-function-decorators/)。 `numpy.testing`模块具有许多装饰器。 下表显示了`numpy.testing`模块中的不同装饰器：

| 装饰器 | 描述 |
| --- | --- |
| `numpy.testing.decorators.deprecated` | 运行测试时，此函数过滤弃用警告 |
| `numpy.testing.decorators.knownfailureif` | 此函数基于条件引发`KnownFailureTest`异常 |
| `numpy.testing.decorators.setastest` | 此装饰器标记测试函数或未测试函数 |
| `numpy.testing.decorators.skipif` | 此函数根据条件引发一个`SkipTest`异常 |
| `numpy.testing.decorators.slow` | 此函数将测试函数或方法标记为缓慢 |

另外，我们可以调用`decorate_methods()`函数将修饰符应用于与正则表达式或字符串匹配的类的方法。

## 实战时间 – 装饰测试函数

我们将直接将`@setastest`装饰器应用于测试函数。 然后，我们将相同的装饰器应用于方法以将其禁用。 另外，我们将跳过其中一项测试，并通过另一项测试。 首先，安装`nose`以防万一。

1.  用`setuptools`安装`nose`：

    ```py
    $ [sudo] easy_install nose

    ```

    或点子：

    ```py
    $ [sudo] pip install nose

    ```

2.  将一个函数当作测试，将另一个函数当作不是测试：

    ```py
    @setastest(False)
    def test_false():
       pass

    @setastest(True)
    def test_true():
       pass
    ```

3.  使用`@skipif`装饰器跳过测试。 让我们使用一个总是导致测试被跳过的条件：

    ```py
    @skipif(True)
    def test_skip():
       pass
    ```

4.  添加一个始终通过的测试函数。 然后，使用`@knownfailureif`装饰器对其进行装饰，以使测试始终失败：

    ```py
    @knownfailureif(True)
    def test_alwaysfail():
         pass
    ```

5.  使用通常应由`nose`执行的方法定义一些`test`类：

    ```py
    class TestClass():
       def test_true2(self):
          pass

    class TestClass2():
       def test_false2(self):
          pass
    ```

6.  让我们从上一步中禁用第二个测试方法：

    ```py
    decorate_methods(TestClass2, setastest(False), 'test_false2')
    ```

7.  使用以下命令运行测试：

    ```py
    $ nosetests -v decorator_setastest.py
    decorator_setastest.TestClass.test_true2 ... ok
    decorator_setastest.test_true ... ok
    decorator_test.test_skip ... SKIP: Skipping test: test_skipTest skipped due to test condition
    decorator_test.test_alwaysfail ... ERROR

    ======================================================================
    ERROR: decorator_test.test_alwaysfail
    ----------------------------------------------------------------------
    Traceback (most recent call last):
     File "…/nose/case.py", line 197, in runTest
     self.test(*self.arg)
     File …/numpy/testing/decorators.py", line 213, in knownfailer
     raise KnownFailureTest(msg)
    KnownFailureTest: Test skipped due to known failure

    ----------------------------------------------------------------------
    Ran 4 tests in 0.001s

    FAILED (SKIP=1, errors=1)

    ```

### 刚刚发生了什么？

我们将某些函数和方法修饰为非测试形式，以便它们被鼻子忽略。 我们跳过了一项测试，也没有通过另一项测试。 我们通过直接使用装饰器并使用`decorate_methods()`函数（请参见`decorator_test.py`）来完成此操作：

```py
from numpy.testing.decorators import setastest
from numpy.testing.decorators import skipif
from numpy.testing.decorators import knownfailureif
from numpy.testing import decorate_methods

@setastest(False)
def test_false():
   pass

@setastest(True)
def test_true():
   pass

@skipif(True)
def test_skip():
   pass

@knownfailureif(True)
def test_alwaysfail():
     pass

class TestClass():
   def test_true2(self):
      pass

class TestClass2():
   def test_false2(self):
      pass

decorate_methods(TestClass2, setastest(False), 'test_false2')
```

## 文档字符串

**Doctests** 是嵌入在 Python 代码中的字符串，类似于交互式会话。 这些字符串可用于测试某些假设或仅提供示例。 `numpy.testing`模块具有运行这些测试的函数。

## 实战时间 – 执行文档测试

让我们写一个简单示例，该示例应该计算众所周知的阶乘，但并不涵盖所有可能的边界条件。 换句话说，某些测试将失败。

1.  `docstring`看起来像您在 Python Shell 中看到的文本（包括提示）。 吊装其中一项测试失败，只是为了看看会发生什么：

    ```py
    """
    Test for the factorial of 3 that should pass.
    >>> factorial(3)
    6
    Test for the factorial of 0 that should fail.
    >>> factorial(0)
    1
    """

    ```

2.  编写以下 NumPy 代码行：

    ```py
    return np.arange(1, n+1).cumprod()[-1]
    ```

    我们希望此代码不时出于演示目的而失败。

3.  例如，通过在 Python Shell 中调用`numpy.testing`模块的`rundocs()`函数来运行`doctest`：

    ```py
    >>> from numpy.testing import rundocs
    >>> rundocs('docstringtest.py')
    Traceback (most recent call last):
     File "<stdin>", line 1, in <module>
     File "…/numpy/testing/utils.py", line 998, in rundocs
     raise AssertionError("Some doctests failed:\n%s" % "\n".join(msg))
    AssertionError: Some doctests failed:
    **********************************************************************
    File "docstringtest.py", line 10, in docstringtest.factorial
    Failed example:
     factorial(0)
    Exception raised:
     Traceback (most recent call last):
     File "…/doctest.py", line 1254, in __run
     compileflags, 1) in test.globs
     File "<doctest docstringtest.factorial[1]>", line 1, in <module>
     factorial(0)
     File "docstringtest.py", line 13, in factorial
     return np.arange(1, n+1).cumprod()[-1]
     IndexError: index -1 is out of bounds for axis 0 with size 0

    ```

### 刚刚发生了什么？

我们编写了文档字符串测试，该测试未考虑`0`和负数。 我们使用`numpy.testing`模块中的`rundocs()`函数运行了测试，结果得到了索引错误（请参见`docstringtest.py`）：

```py
import numpy as np

def factorial(n):
   """
   Test for the factorial of 3 that should pass.
   >>> factorial(3)
   6

   Test for the factorial of 0 that should fail.
   >>> factorial(0)
   1
   """
   return np.arange(1, n+1).cumprod()[-1]
```

## 总结

您在本章中了解了测试和 NumPy 测试工具。 我们介绍了单元测试，文档字符串测试，断言函数和浮点精度。 大多数 NumPy 断言函数都会处理浮点数的复杂性。 我们展示了可以被鼻子使用的 NumPy 装饰器。 装饰器使测试更加容易，并记录了开发人员的意图。

下一章的主题是 matplotlib -- Python 科学的可视化和图形化开源库。